#include <gtest/gtest.h>

#include <algorithm>

#include "simulateur.h"

namespace six_qui_prend::core {

TEST(choisir_carte, strategie_kCarte_la_plus_petite) {
  Joueuse j{
    strategie_t::kCarte_la_plus_petite,
    {2, 4, 6},
    0,
    {},
    0
  };

  choisir_carte(j);

  EXPECT_EQ(cartes_t({4, 6}), j.main);
  EXPECT_EQ(2, j.carte_pour_ce_tour);
}    
  
TEST(choisir_carte, strategie_kCarte_au_milieu_main_quand_main_a_nombre_cartes_impair) {
  Joueuse j{
    strategie_t::kCarte_au_milieu_main,
    {2, 4, 6},
    0,
    {},
    0
  };

  choisir_carte(j);

  EXPECT_EQ(cartes_t({2, 6}), j.main);
  EXPECT_EQ(4, j.carte_pour_ce_tour);
}

TEST(choisir_carte, strategie_kCarte_au_milieu_main_quand_main_a_nombre_cartes_pair) {
  Joueuse j{
    strategie_t::kCarte_au_milieu_main,
    {2, 4},
    0,
    {},
    0
  };

  choisir_carte(j);

  EXPECT_EQ(cartes_t({2}), j.main);
  EXPECT_EQ(4, j.carte_pour_ce_tour);
}

TEST(choisir_carte, strategie_kCarte_random) {
  cartes_t main_reference{2, 4, 6};
  Joueuse j{
    strategie_t::kCarte_random,
    main_reference,
    0,
    {},
    0
  };

  choisir_carte(j);

  // Pour la strategie random, tout ce qu'on peut verifier est que carte_pour_ce_tour
  // etait bien dans main et qu'elle a ete retiree de j.main
  EXPECT_TRUE(std::ranges::find(main_reference, j.carte_pour_ce_tour) != main_reference.end());

  cartes_t main_apres{main_reference};
  erase_if(main_apres, [&j](const auto &carte) { return carte == j.carte_pour_ce_tour; });
  EXPECT_EQ(main_apres, j.main);
}

TEST(choisir_carte, strategie_kCarte_avant_la_plus_grande_main_a_au_moins_2_cartes) {
  Joueuse j{
    strategie_t::kCarte_avant_la_plus_grande,
    {2, 4, 6},
    0,
    {},
    0
  };

  choisir_carte(j);

  EXPECT_EQ(cartes_t({2, 6}), j.main);
  EXPECT_EQ(4, j.carte_pour_ce_tour);
}

TEST(choisir_carte, strategie_kCarte_avant_la_plus_grande_main_a_une_seule_carte) {
  Joueuse j{
    strategie_t::kCarte_avant_la_plus_grande,
    {2},
    0,
    {},
    0
  };

  choisir_carte(j);

  EXPECT_EQ(cartes_t({}), j.main);
  EXPECT_EQ(2, j.carte_pour_ce_tour);
}

TEST(choisir_carte, strategie_kCarte_la_plus_grande) {
  // TODO 10 bis
  EXPECT_TRUE(false);
}

TEST(get_nb_vaches, carte_simple) {
  EXPECT_EQ(1, get_nb_vaches(1));
  EXPECT_EQ(2, get_nb_vaches(5));
  EXPECT_EQ(3, get_nb_vaches(10));
  EXPECT_EQ(5, get_nb_vaches(11));
  EXPECT_EQ(7, get_nb_vaches(55));
  EXPECT_EQ(1, get_nb_vaches(104));
}

TEST(get_nb_vaches, cartes) {
  const cartes_t cartes{1, 15, 20, 44, 55, 104};
  EXPECT_EQ(19, get_nb_vaches(cartes));
}

TEST(get_rang_ou_jouer, la_carte_peut_etre_ajoutee_a_une_rangee) {
  const Joueuse j{
    strategie_t::kCarte_la_plus_grande,
    {2, 4, 6},
    42,
    {},
    0
  };
  const rangees_t rangees{
    {6, 7, 8},
    {38, 39, 40},
    {50, 60},
    {103, 104}
  };

  EXPECT_EQ(1, get_rang_ou_jouer(j, rangees));
}

TEST(get_rang_ou_jouer, la_carte_est_trop_petite_et_doit_se_substituer_a_la_rangee_qui_rapporte_le_moins_de_vaches) {
  const Joueuse j{
    strategie_t::kCarte_la_plus_grande,
    {2, 4, 6},
    42,
    {},
    0
  };
  const rangees_t rangees{
      {6, 7, 8, 98}, // Nombre de vaches = 4
      {38, 39, 90},  // Nombre de vaches = 5
      {50, 60}, // Nombre de vaches = 6
      {103, 104} // Nombre de vaches = 2
  };

  EXPECT_EQ(3, get_rang_ou_jouer(j, rangees));
}

TEST(jouer_carte, la_carte_est_ajoutee_a_une_rangee_de_moins_de_5_cartes) {
  const Joueuse j_init{
    strategie_t::kCarte_la_plus_grande,
    {2, 4, 6},
    42,
    {12, 13},
    2
  };
  Joueuse j{j_init};
  const rangees_t rangees_init{
      {6, 7, 8},
      {38, 39, 40},
      {50, 60},
      {103, 104}
  };
  rangees_t rangees{rangees_init};

  jouer_carte(j, rangees);

  EXPECT_EQ(j_init.main, j.main);
  EXPECT_EQ(j_init.cartes_encaissees, j.cartes_encaissees);
  EXPECT_EQ(j_init.nb_vaches_encaissees, j.nb_vaches_encaissees);
  EXPECT_EQ(rangees_init[0], rangees[0]);
  EXPECT_EQ(cartes_t({38, 39, 40, 42}), rangees[1]);
  EXPECT_EQ(rangees_init[2], rangees[2]);
  EXPECT_EQ(rangees_init[3], rangees[3]);
}

  TEST(jouer_carte, la_carte_est_ajoutee_a_une_rangee_de_5_cartes) {
  const Joueuse j_init{
    strategie_t::kCarte_la_plus_grande,
    {2, 4, 6},
    42,
    {12, 13},
    2
  };
  Joueuse j{j_init};
  const rangees_t rangees_init{
        {6, 7, 8},
        {35, 36, 38, 39, 40},
        {50, 60},
        {103, 104}
  };
  rangees_t rangees{rangees_init};

  jouer_carte(j, rangees);

  EXPECT_EQ(j_init.main, j.main);
  EXPECT_EQ(cartes_t({12, 13, 35, 36, 38, 39, 40}), j.cartes_encaissees);
  EXPECT_EQ(10, j.nb_vaches_encaissees);
  EXPECT_EQ(rangees_init[0], rangees[0]);
  EXPECT_EQ(cartes_t({42}), rangees[1]);
  EXPECT_EQ(rangees_init[2], rangees[2]);
  EXPECT_EQ(rangees_init[3], rangees[3]);
}

  TEST(jouer_carte, la_carte_se_substitue_a_une_rangee_de_cartes) {
  const Joueuse j_init{
    strategie_t::kCarte_la_plus_grande,
    {2, 4, 6},
    42,
    {12, 13},
    2
  };
  Joueuse j{j_init};
  const rangees_t rangees_init{
        {6, 7, 8, 98}, // Nombre de vaches = 4
        {38, 39, 90},  // Nombre de vaches = 5
        {50, 60}, // Nombre de vaches = 6
        {103, 104} // Nombre de vaches = 2
  };
  rangees_t rangees{rangees_init};

  jouer_carte(j, rangees);

  EXPECT_EQ(j_init.main, j.main);
  EXPECT_EQ(cartes_t({12, 13, 103, 104}), j.cartes_encaissees);
  EXPECT_EQ(4, j.nb_vaches_encaissees);
  EXPECT_EQ(rangees_init[0], rangees[0]);
  EXPECT_EQ(rangees_init[1], rangees[1]);
  EXPECT_EQ(rangees_init[2], rangees[2]);
  EXPECT_EQ(cartes_t({42}), rangees[3]);
}
} // six_qui_prend::core
